############################################################
## AM can be configured to run in a number of different
## ways.  The configuration info is in the file AM.config.
## In previous versions of amc.pl, this file was read, and
## then numerous variables were set to change how the
## program ran.  This entailed repeated variable
## interpolation and flag checking in many loop constructs,
## potentially slowing down execution.
##
## In this new version, the config file is read and parsed;
## then, based on the configuration, the code for the
## various Perl subroutines called from amc.c are first
## written as "personalized" strings and then evaluated
## using eval().
##
## Hopefully you can see the speed-up in running time.  If
## not -- well, I feel better anyway.
############################################################

############################################################
sub setup {

  ##srand(time() ^ ($$ + ($$ << 15)));
  srand(0);

## set input record separator for slurping

  undef $/;

## config file

  open (CONFIG, "AM.config");
  $_ = <CONFIG>;
  close CONFIG;

  ($name) = /Project Name.*:\s*(.*)/;

  ($excNull) = /null.*:\s*(.*)/;
  $excNull = 0 unless $excNull eq 'exclude';

  ($excGiven) = /given.*:\s*(.*)/;
  $excGiven = 0 unless $excGiven eq 'exclude';

  ($probability) = /probability.*:\s*(.*)/;
  ($repeat) = /repeat.*:\s*(.*)/;
  ($collate) = /collate.*:\s*(.*)/;
  $collate = 0 unless $collate eq 'yes' and $repeat > 1;

  ($linearOrSq) = /linear or squared.*:\s*(.*)/;

  ($outputToFile) = /output to file.*:\s*(.*)/;
  if ($outputToFile eq 'yes') {
    open (PFLE, ">$name/newamcresults");
    select PFLE;
  }

  ($skipset) = /skip analogical set.*:\s*(.*)/;
  $skipset = 0 unless $skipset eq 'yes';

  ($useOutcomeFile) = /use \.outcome file.*:\s*(.*)/;

  ($specifyFreq) = /specify frequency.*:\s*(.*)/;
  $specifyFreq = 0 if $specifyFreq eq 'no';

  ($usescommas) = /format with commas.*:\s*(.*)/;

  ($winloss) =    /show win-loss.*:\s*(.*)/;
  ($predictive) = /show predictive success.*:\s*(.*)/;
  ($confusion) =  /show confusion.*:\s*(.*)/;
  $winloss = (($winloss eq 'yes') or
              ($predictive eq 'yes') or
	      ($confusion eq 'yes')) ? 1 : 0;

## read the routines in

  my($setup, $subs);
  $setup = <DATA>;

## reset input record separator

  $/ = "\n";

## adjust for comma formatting

  if ($usescommas eq 'yes') {
    $setup =~ s'_\$bigsep_'\s*,\s*'g; # assume comma-separated
    $setup =~ s'_\$smallsep_'\s+'g;   # and space-separated
  } else {
    $setup =~ s'_\$bigsep_'\s+'g;     # assume space-separated
    $setup =~ s'_\$smallsep_''g;      # and not separated
  }

## split setup routine from subs

  ($setup, $subs) = split /## BREAK ##/, $setup;

## read the outcome, data, and test files

  eval ($setup);

## adjust the definitions of the subs

  $subs =~ s'_@data_'scalar @data'eg;

  if ($excNull) {
    $subs =~ s/## begin include null.*?## end include null//sg;
    $subs =~ s/Nulls: include\n//s;
  } else {
    $subs =~ s/## begin exclude null.*?## end exclude null//sg;
    $subs =~ s/Nulls: exclude\n//s;
  }

  if ($excGiven) {
    $subs =~ s/## begin include given.*?## end include given//sg;
    $subs =~ s/Include context even if it is in the data file\n//s;
  } else {
    $subs =~ s/## begin exclude given.*?## end exclude given//sg;
    $subs =~ s/If context is in data file then exclude\n//s;
 }

  $subs =~ s'_\$probability_'$probability'eg;
  if ($probability == 1) {
    $subs =~ s/## begin probability.*?## end probability//sg;
  }

  if (!$collate) {
    $subs =~ s/## begin collate.*?## end collate//sg;
  }

  $subs =~ s'_\$linearOrSq_'$linearOrSq'eg;
  if ($linearOrSq eq 'linear') {
    $subs =~ s/## begin squared.*?## end squared//sg;
  } else {
    $subs =~ s/## begin linear.*?## end linear//sg;
  }

  if ($skipset) {
    $subs =~ s/## begin analogical set.*?## end analogical set//sg;
  }

  if (!$winloss) {
    $subs =~ s/## begin win-loss.*?## end win-loss//sg;
  }

## define the subs

  eval ($subs);

}

__DATA__

############################################################
## Now come two routines.  The first parses the outcome,
## data, and test files.  The second contains the
## subroutines called from amc.c.  They are separated by the
## line containing the word BREAK.

## data file

open(DATA, "$name/data");
while(<DATA>) {
  chomp;
  my($outcome, $data, $spec) = split /_$bigsep_/, $_, 3;

  push @outcome, $outcome;
  push @spec, $spec;
  push @data, [split /_$smallsep_/, $data];

  last if $specifyFreq and ($specifyFreq == @data);

}
close DATA;

## outcome file

@outcomelist = @ocl = (undef);
my $outcomecounter = 0;
if (-e "$name/outcome") {
  open(OUTCOME, "$name/outcome");
  while(<OUTCOME>) {
    chomp;
    my($oc, $outcome) = split /\s+/, $_, 2;
    $outcometonum{$oc} = ++$outcomecounter;
    push @outcomelist, $outcome;
    push @ocl, $oc;
  }
  close OUTCOME;
} else {
  my %oc = ();
  map { ++$oc{$_} } @outcome;
  foreach (sort {lc($a) cmp lc($b)} keys %oc) {
    $outcometonum{$_} = ++$outcomecounter;
    push @outcomelist, $_;
    push @ocl, $_;
  }
}
@outcome = map { $outcometonum{$_} } @outcome;

## test file

open(TEST, "$name/test");
@testItems = <TEST>;
close TEST;
chomp(@testItems);
(undef, $item) = split /_$bigsep_/, $testItems[0];
$maxvar = scalar split /_$smallsep_/, $item;
$totalTestItemRuns = $repeat * @testItems;
$totalCorrectPredictions = 0;

## BREAK ##

############################################################
sub parseTestItem {

## parse test item

## begin collate
  $validTestItemRuns = 0;
  @testPct = ();
  @testPctSq = ();
  $correctPredictions = 0;
## end collate

  my $item;
  ($curTestOutcome, $item, $curTestSpec)
    = split /_$bigsep_/, (shift @testItems), 3;
  $curTestOutcome = $outcometonum{$curTestOutcome};

## begin exclude null
  my $eq = $item =~ tr/=/=/;
  @curTestItem = split /_$smallsep_/, $item;
  $activeVar = @curTestItem - $eq;
## end exclude null
## begin include null
  @curTestItem = split /_$smallsep_/, $item;
  $activeVar = @curTestItem;
## end include null
  $activeContexts = 1 << $activeVar;
}

############################################################
sub beginTestItem {

## initialize count

  @datacount = @sum = ();
  $grandtotal = 0;
  $excludedData = 0;

## begin probability
  $probvec = pack "b*", "1" x _@data_;
  for (my $i = 0; $i < _@data_; ++$i) {
    vec($probvec, $i, 1) = 0, ++$excludedData
      if rand() > $probability;
  }
## end probability

## print header

  my ($sec,$min,$hour) = localtime();
  print "Time: ${hour}h ${min}m ${sec}s\n\n";
  print STDOUT "Time: ${hour}h ${min}m ${sec}s\n\n";
  print STDOUT "@curTestItem\n";

## compare data and given

  %subcontext = ();
  @level = ();
  %subtooutcome = ();
  my $testindata = 0;
  my $eg = 0;
  my $i;
  for ($i = 0; $i < _@data_; $i++) {
## begin probability
    next unless vec($probvec, $i, 1);
## end probability
    my @dataItem = @{$data[$i]};
    my $context = 0;
    my $l = 0;
    my $j;
    for ($j = 0; $j < @curTestItem; $j++) {
## begin exclude null
      next if $curTestItem[$j] eq '=';
## end exclude null
      my $bit = $curTestItem[$j] ne $dataItem[$j];
      $context = ($context << 1) | $bit;
      $l += $bit;
    }
    push @{$subcontext{$context}}, $i;
    ++$level[$l]{$context};
    my $outcome = $outcome[$i];
    next if ($subtooutcome{$context} == $outcome);
    $subtooutcome{$context} = $outcome, next
      unless defined $subtooutcome{$context};
    $subtooutcome{$context} = 0;
  }

  if ($level[0]{0}) {
    ++$testindata;
## begin exclude given
    delete $level[0]{0}, ++$eg if $excGiven;
## end exclude given
  }
  my(@subcontexts) = ();
  @level = reverse @level;
  foreach (@level) {
    push @subcontexts, keys %{$_};
  }
  $numsubcontexts = scalar @subcontexts;
  $subcontexts = pack "L*", @subcontexts;
  $suboutcomes = pack "C*", map { $subtooutcome{$_} } @subcontexts;

## We use a here document to print the header.
## Unwanted lines are removed when sub setup is run

  print <<TOP;
Project Name: $name
Given Context:  @curTestItem, $curTestSpec
If context is in data file then exclude
Include context even if it is in the data file
Number of data points: _@data_
Probability of including any one data point: _$probability_
Total Excluded: $excludedData @{[ $eg ? " + test item" : "" ]}
Nulls: exclude
Nulls: include
Gang: _$linearOrSq_
Number of active variables: $activeVar
Number of active contexts: $activeContexts
TOP
  print "Test item is in the data.\n" if $testindata;
  print "\n\n";
}

############################################################
sub count {
  my $count = $_[0];
  @list = ();
  foreach (unpack "L*", $_[1]) {
    push @list, @{$subcontext{$_}};
  }
  foreach (@list) {
## begin squared
    $datacount[$_] += $count * @list;
    $sum[$outcome[$_]] += $count * @list;
    $grandtotal += $count * @list;
## end squared
## begin linear
    $datacount[$_] += $count;
    $sum[$outcome[$_]] += $count;
    $grandtotal += $count;
## end linear
  }
}

############################################################
sub endTestItem {

## begin analogical set
  if ($grandtotal) {
    print "Analogical Set\n\n";
    print "Total Frequency = $grandtotal\n\n";
    my $i;
    for ($i = 0; $i < @datacount; $i++) {
      next unless $datacount[$i];
      printf "%-9.9s\t%-11.11s\t%7.0f\t%5.2f%%\n",
        $ocl[$outcome[$i]],
        $spec[$i],
        $datacount[$i],
        100 * $datacount[$i]/$grandtotal;
    }
    print "\n\n";
  }
## end analogical set

  if (!$grandtotal) {
    print "No data items considered.  No prediction possible.\n\n\n";
## begin win-loss
    ++$winloss[$curTestOutcome][0];
## end win-loss
  } else {
    print "Statistical Summary\n\n";
## begin collate
    ++$validTestItemRuns;
## end collate
    my $high = 0;
    for ($i = 1; $i < @outcomelist; ++$i) {
      my $n;
      next unless $n = $sum[$i];
      $high = $n if $n > $high;
      printf "%-9.9s\t%-11.11s\t%7.0f\t%5.2f%%\n",
        $ocl[$i],
        $outcomelist[$i],
        $n,
        100 * $n/$grandtotal;
## begin collate
      $testPct[$i] += $n/$grandtotal;
      $testPctSq[$i] += ($n*$n)/($grandtotal*$grandtotal);
## end collate
    }
    if ($sum[$curTestOutcome] == $high) {
      print "\nCorrect outcome predicted.\n";
      ++$totalCorrectPredictions;
## begin collate
      ++$correctPredictions;
## end collate
## begin win-loss
      ++$winloss[$curTestOutcome][$curTestOutcome];
## end win-loss
    } else {
      print "\nIncorrect outcome predicted.\n";
## begin win-loss
      my @winners = ();
      my $i;
      for ($i = 1; $i < @outcomelist; ++$i) {
	push @winners, $i if $sum[$i] == $high;
      }
      my $numwinners = scalar @winners;
      foreach (@winners) {
	$winloss[$curTestOutcome][$_] += 1 / $numwinners;
      }
## end win-loss
    }
    print "\n\n";
  }
}
############################################################
## subroutine collate

sub collate {
## begin collate
  print "Summary for this test item:\n";
  print "Number of runs used: $validTestItemRuns out of $repeat\n\n";
  print "\n" and return unless $validTestItemRuns;
  print "         \t           \t Avg  \tStd Dev\n";
  my $i;
  for ($i = 1; $i < @outcomelist; ++$i) {
    next unless $testPct[$i];
    if ($validTestItemRuns > 1) {
      printf "%-9.9s\t%-11.11s\t%5.2f%%\t %5.2f%%\n",
        $ocl[$i],
        $outcomelist[$i],
        100 * $testPct[$i] / $validTestItemRuns,
        100*sqrt(($testPctSq[$i]-$testPct[$i]*$testPct[$i]/$validTestItemRuns)
		 /($validTestItemRuns-1));
    } else {
      printf "%-9.9s\t%-11.11s\t%5.2f%%\n",
        $ocl[$i],
        $outcomelist[$i],
        100 * $testPct[$i] / $validTestItemRuns;
    }
  }
  printf "\nCorrect prediction occurred %5.2f%% (%i/%i) of the time\n",
    100 * $correctPredictions / $repeat,
    $correctPredictions,
    $repeat;
  print "\n\n";
## end collate
}

############################################################
sub closeshop {
  printf "Over correct prediction occurred %5.2f%% (%i/%i) of the time\n\n",
    100 * $totalCorrectPredictions / $totalTestItemRuns,
    $totalCorrectPredictions,
    $totalTestItemRuns;

## begin win-loss
  my($i,$j);
  for ($i = 1; $i < @outcomelist; ++$i) {
    my $total = 0;
    foreach (@{$winloss[$i]}) {
      $total += $_;
    }
    next unless $total;
    print "Test items with outcome $outcomelist[$i] were predicted as follows:\n";
    for ($j = 1; $j < @outcomelist; ++$j) {
      my $t;
      next unless ($t = $winloss[$i][$j]);
      printf "%5.2f%% %s\t(%i/%i)\n", 100 * $t / $total, $outcomelist[$j], $t, $total;
    }
    if ($t = $winloss[$i][0]) {
      printf "%5.2f%% could not be predicted (%i/%i)\n", 100 * $t / $total, $t, $total;
    }
      
    print "\n\n";
  }
## end win-loss

  my ($sec,$min,$hour) = localtime();
  print "Time: ${hour}h ${min}m ${sec}s\n\n";
  print STDOUT "Time: ${hour}h ${min}m ${sec}s\n\n";

  if ($outputToFile eq 'yes') {
    close PFLE;
    print STDOUT "File is done.  Look at $name/newamcresults\n";
  }
}
